/*-
 * Copyright (C)2007..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2007..2019 @BABOLO http://www.babolo.ru/\n"
#ident "@(#) $Id: events.c,v 1.13 2021/08/28 13:43:26 babolo Exp $\n"

#define BLIN_COMPAT  4
#define MIFE_COMPAT  5
#define Bpars_COMPAT 4
#define MULAR_COMPAT 0

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "popushd.h"
#include "react.h"

extern char    **environ;
extern u_int32_t popush_state;

void
popush_cnfree(popush_cnf *cfg) {
    if  (cfg) {
        if  (cfg->rawobj)   mular_destroy(cfg->rawobj);
        if  (cfg->reaction) mular_destroy(cfg->reaction);
        if  (cfg->cluster)  mular_destroy(cfg->cluster);
        if  (cfg->object)   mular_destroy(cfg->object);
        if  (cfg->ask)      mular_destroy(cfg->ask);
        free(cfg);
        cfg = NULL;
}   }

int
popush_diffmerge(popush_main *main, popush_cnf *cfg) {
#   define blin_internal_flags (main->flags & BLIN_MASK)
    int             ex = EX_OK;
    struct timespec noll;
    struct stat     ost;
    struct stat     nst;
    size_t          opo;
    size_t          npo;
    popush_obj     *ooo;
    popush_obj     *noo;

    ifBLIN_QX3("+");
    bzero(&noll, sizeof(noll));
    cfg->konfik.evnt.fflags = NOTE_LINK|NOTE_DELETE|NOTE_WRITE|NOTE_EXTEND;
    cfg->konfik.evnt.fflags |= NOTE_RENAME|NOTE_REVOKE|NOTE_ATTRIB;
    cfg->konfik.evnt.filter = EVFILT_VNODE;
    if  (main->cnf) {
        /*       ,
         *        
         *    
         */
        if  (main->cnf->self) cfg->konfik.evnt.flags = EV_ADD | EV_CLEAR;
        for (opo = 0; opo < MULAR_NEXT(main->cnf->object); opo++) {
            int found, f;

        ;   if  (!(ooo = mular_getix(main->cnf->object, opo))) {
        ;       ifBLIN_QW2("mular_getix old object %u", opo);
        ;       continue;
        ;   }
        ;   if  (!(ooo->evnt.flags)) continue;
        ;   for (found = 0, npo = 0; npo < MULAR_NEXT(cfg->object); npo++) {
        ;       if  (!(noo = mular_getix(cfg->object, npo))) {
        ;           ifBLIN_QW2("mular_getix new object");
        ;           continue;
        ;       }
        ;       if  (strcmp(noo->fname, ooo->fname) == 0) {
        ;           found = 1;
        ;           if  (fstat(ooo->evnt.ident, &ost) < 0) {
        ;               ifBLIN_QW1("fstat old %s", ooo->fname);
        ;               break;
        ;           }
        ;           if  ((f = open(noo->fname, O_RDONLY)) < 0) {
        ;               ifBLIN_QW1("open #1 %s", noo->fname);
        ;               break;
        ;           }
        ;           if  (fstat(f, &nst) < 0) {
        ;               ifBLIN_QW1("fstat new %s", noo->fname);
        ;               close(f);
        ;               break;
        ;           }
        ;           if  ((ost.st_dev == nst.st_dev) && (ost.st_ino == nst.st_ino)) {
        ;               close(f);
        ;               noo->evnt.ident = ooo->evnt.ident;
        ;               ooo->evnt.fflags &= ~noo->evnt.fflags;
        ;               ooo->evnt.flags = EV_DELETE;
        ;               if  (ooo->evnt.fflags && (kevent(main->kq, &(ooo->evnt), 1, NULL, 0, &noll) < 0)) {
        ;                   ifBLIN_QW1("kevent #1 for %s", ooo->fname);
        ;               }
        ;               ooo->evnt.flags = 0;
        ;           } else {
        ;               fcntl(f, F_SETFD, FD_CLOEXEC);
        ;               close(ooo->evnt.ident);
        ;               ooo->evnt.flags = 0;
        ;               noo->evnt.ident = f;
        ;           }
        ;           noo->evnt.flags = EV_ADD | EV_CLEAR;
        ;           if  (kevent(main->kq, &(noo->evnt), 1, NULL, 0, &noll) < 0) {
        ;               ifBLIN_QW1("kevent #2 for %s", noo->fname);
        ;               noo->evnt.flags = 0;
        ;   }   }   }
        ;   if  (!found) {
        ;       close(ooo->evnt.ident);
        ;       ooo->evnt.flags = 0;
    }   }   }
    if  (cfg->self && !cfg->konfik.evnt.flags) {
        cfg->konfik.evnt.flags = EV_ADD | EV_CLEAR;
        if  (kevent(main->kq, &(cfg->konfik.evnt), 1, NULL, 0, &noll) < 0) {
            ifBLIN_QW1("kevent #3 for %s", main->cnfname);
            cfg->konfik.evnt.flags = 0;
    }   }
    for (npo = 0; npo < MULAR_NEXT(cfg->object); npo++) {
        if  (!(noo = mular_getix(cfg->object, npo))) {
            ifBLIN_QW2("mular_getix new object %u", npo);
            continue;
        }
        if  (noo->evnt.flags) continue;
        noo->evnt.flags = EV_ADD | EV_CLEAR;
        if  ((int)(noo->evnt.ident = open(noo->fname, O_RDONLY)) < 0) {
            ifBLIN_QW2("open #2 %s", noo->fname);
            noo->evnt.flags = 0;
            continue;
        }
        if  (kevent(main->kq, &(noo->evnt), 1, NULL, 0, &noll) < 0) {
            ifBLIN_QW2("kevent #3 for %s", noo->fname);
            noo->evnt.flags = 0;
            continue;
    }   }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
popush_setkact(popush_main *main, struct kevent *ev) {
#   define blin_internal_flags (main->flags & BLIN_MASK)
    int ex = EX_OK;

    ifBLIN_QX3("+");
    if  (ev->flags & EV_ERROR) {
        ifBLIN_QX1("EV_ERROR %"BLIN_U, ev->data);
        switch (ev->data) {
        case EINTR:
            popush_setaction(main, main->cnf->sign);
            break;
        case EFAULT:
            /* FALLTHROUGH */
        case ENOMEM:
            ex = EX_TEMPFAIL;
            break;
        }
    } else if (ev->filter == EVFILT_SIGNAL) {
        ifBLIN_QX2("EVFILT_SIGNAL %"BLIN_U, ev->data);
        popush_setaction(main, main->cnf->sign);
    } else if (ev->filter == EVFILT_VNODE) {
        int         found;
        size_t      n;
        popush_obj *o;

        ifBLIN_QX2("EVFILT_VNODE %"BLIN_U" %04X", ev->data, ev->fflags);
        if  (main->cnf->konfik.evnt.ident == ev->ident) {
            popush_state |= POPUSH_FSELF;
            found = 1;
        } else {
            found = 0;
        }
        for (n = 0; !found && n < MULAR_NEXT(main->cnf->object); n++) {
            /*    ,
             *      
             * XXXX
             */
            if  (!(o = mular_getix(main->cnf->object, n))) {
                ifBLIN_QW1("NULL object %"BLIN_U, n);
                continue;
            }
            if  (o->evnt.ident == ev->ident) {
                if  (ev->fflags & NOTE_DELETE) popush_setaction(main, o->fdel);
                if  (ev->fflags & NOTE_WRITE ) popush_setaction(main, o->fwrt);
                if  (ev->fflags & NOTE_EXTEND) popush_setaction(main, o->fext);
                if  (ev->fflags & NOTE_ATTRIB) popush_setaction(main, o->fatr);
                if  (ev->fflags & NOTE_LINK  ) popush_setaction(main, o->flnk);
                if  (ev->fflags & NOTE_RENAME) popush_setaction(main, o->fren);
                if  (ev->fflags & NOTE_REVOKE) popush_setaction(main, o->frev);
                found = 1;
        }   }
        if  (!found) ifBLIN_QX0("Not found fd %"BLIN_D, ev->ident);
    } else {
        ifBLIN_QW0( "Not expectative event { %"BLIN_U" %d %04X %08X %"BLIN_U" %"BLIN_X" }"
                  , ev->ident
                  , ev->filter
                  , ev->flags
                  , ev->fflags
                  , ev->data
                  , BLIN_I(ev->udata)
                  );
    }
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
popush_setevent(popush_main *main) {
#   define blin_internal_flags (main->flags & BLIN_MASK)
    int             ex = EX_OK;
    struct kevent   evnt;
    struct timespec noll;
    int             cfg;

    ifBLIN_QX3("+");
    bzero(&noll, sizeof(noll));
    bzero(&evnt, sizeof(evnt));
    while (!ex && ((cfg = kevent(main->kq, NULL, 0, &evnt, 1, &noll)) > 0)) {
        ifBLIN_QX2( "kevent{ %"BLIN_U" %d %04X %08X %"BLIN_U" %"BLIN_X" }"
                  , evnt.ident
                  , evnt.filter
                  , evnt.flags
                  , evnt.fflags
                  , evnt.data
                  ,  BLIN_I(evnt.udata)
                  );
        ex = popush_setkact(main, &evnt);
    }
    ifBLIN_QX3("-");
    return(ex);
#   undef blin_internal_flags
}

static int
vforkexec(char **argv) {
    int pid;

    pid = vfork();
    if  (!pid) execve(*argv, argv, environ);
    return(pid);
}

void
popush_execevent(popush_main *main) {
#   define blin_internal_flags (main->flags & BLIN_MASK)
    int           status;
    pid_t         pid;
    popush_clust *c;
    babolo_parm **p;
    size_t        n;

    ifBLIN_QX3("+");
    for (n = 0; n < MULAR_NEXT(main->cnf->cluster); n++) {
        size_t i = n >> 3;
        u_char j = 1 << (n & 7);

        if  (main->actionbit[i] & j) {
            if  (!(c = mular_getix(main->cnf->cluster, n))) {
                ifBLIN_QW0("mular_getix cluster #1 %u", n);
                break;
            }
            switch ((pid = fork())) {
            case -1:
                ifBLIN_QW0("fork %u", n);
                break;
            case 0: {
                size_t q;
                size_t d;
                size_t i;
                int    k;

            ;   q = c->clust;
            ;   d = MULAR_NEXT(main->cnf->cluster);
            ;   if  (n + 1 < d) {
            ;       if  (!(c = mular_getix(main->cnf->cluster, n + 1))) {
            ;           ifBLIN_QW0("mular_getix cluster #2 %u", n);
            ;           break;
            ;       }
            ;       d = c->clust;
            ;   }
            ;   ifBLIN_QX4("%"BLIN_U"E %"BLIN_U" %"BLIN_U, n, q, d);
            ;   for (; q < d; q++) {
            ;       ifBLIN_QX4("%p to %p", q, d);
            ;       if  (!(p = mular_getix(main->cnf->reaction, q))) {
            ;           ifBLIN_QW0("mular_getix reaction #1 %u", q);
            ;           break;
            ;       }
            ;       ifBLIN_QO4 {
            ;           for (i = 0; i < (*p)->argc; i++) {
                            ifBLIN_QX4("%"BLIN_U"[%"BLIN_D"]%s~", n, i, (*p)->argv[i]);
            ;       }   }
            ;       if  (babolo_testword(&react, (*p)->argv[1]) == POPUSH_EFILE) {
            ;           if  ((pid = vforkexec(&((*p)->argv[1]))) < 0) {
                            ifBLIN_QW0("vfork %"BLIN_U" %"BLIN_U, n, q);
            ;           } else if (babolo_testword(&react, (*p)->argv[0]) == POPUSH_EZERO) {
            ;               ifBLIN_QX4("exit v %d %d", pid, waitpid(pid, &status, 0));
            ;           }
            ;       } else if (babolo_testword(&react, (*p)->argv[1]) == POPUSH_EWAIT) {
            ;           for (k = 1; k;) {
            ;               pid = wait(&status);
            ;               ifBLIN_QX4("exit v %d %08X\n", pid, status);
            ;               if  ((pid < 0) && (errno != EINTR)) k = 0;
            ;           }
            ;       } else {
            ;           ifBLIN_QX1("unknown %s~", (*p)->argv[1]);
            ;   }   }
            ;   _exit(0);
            }
            default:
                c->pid = pid;
                ifBLIN_QX4("fork %d", pid);
    }   }   }
    for (n = 0; n < MULAR_NEXT(main->cnf->cluster); n++) {
        if  (!(c = mular_getix(main->cnf->cluster, n))) {
            ifBLIN_QW1("mular_getix cluster #3 %u", n);
        } else if (c->pid) {
            pid = waitpid(c->pid, &status, 0);
            ifBLIN_QX4("exit %d %d", c->pid, pid);
            c->pid = 0;
    }   }
    bzero(main->actionbit, (n + 7) >> 3);
    ifBLIN_QX3("-");
#   undef blin_internal_flags
}
