/*-
 * Copyright (C)2005..2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: popushd.h,v 1.17 2021/08/28 13:43:26 babolo Exp $
 */

#define POPUSH_DMN   0x000001
#define POPUSH_ERX   0x000002
#define POPUSH_CRI   0x000004
#define POPUSH_ANA   0x000008

/*   POPUSH_*      */
/*               */
#define POPUSH_EROR   0x00
#define POPUSH_NULL   0x08
#define POPUSH_FULL   0x08
#define POPUSH_ZERO   0x01
#define POPUSH_ASYN   0x02
#define POPUSH_STRT   0x03
#define POPUSH_STOP   0x04
#define POPUSH_SIGN   0x05
#define POPUSH_SELF   0x06
#define POPUSH_FILE   0x07
#define POPUSH_MAXP   0x10

#define POPUSH_FSTRT   0x01
#define POPUSH_FSTOP   0x02
#define POPUSH_FSIGN   0x04
#define POPUSH_FSELF   0x08

#define POPUSH_EZERO   1
#define POPUSH_EASYN   2
#define POPUSH_EFILE   3
#define POPUSH_EWAIT   4
#define POPUSH_ESIGN   5

typedef struct popush_obj {
    struct kevent evnt;
    char         *fname;
    size_t        fdel;
    size_t        fwrt;
    size_t        fext;
    size_t        fatr;
    size_t        flnk;
    size_t        fren;
    size_t        frev;
} popush_obj;

typedef struct popush_clust {
    size_t      clust;
    pid_t       pid;
} popush_clust;

typedef struct popush_cnf {
    mular_descriptor *rawobj;   /* *babolo_parm  */
    mular_descriptor *reaction; /* *babolo_parm  */
    mular_descriptor *cluster;  /* popush_clust  */
    mular_descriptor *object;   /* popush_obj	 */
    mular_descriptor *ask;      /* u_char bit[]  */
    size_t            strt;
    size_t            stop;
    size_t            sign;
    size_t            self;
    popush_obj        konfik;
} popush_cnf;

typedef struct popush_main {
    u_int32_t        flags;
    u_char           absq;        /*                */
    u_int8_t         dummy1;
    u_int16_t        dummy2;
    popush_cnf      *cnf;         /*                */
    mife_descriptor *cnfile;
    char            *cnfname;     /*      */
    char            *pidname;     /*    PID      */
    char            *wdpath;      /* cwd               */
    u_char          *actionbit;
    int              kq;          /*   kqueue           */
} popush_main;

/*
 *       2 multilar:
 *  - rawobj,    babolo_parm     ,
 *         ,    
 *       babolo_parm;
 *  - reaction,    babolo_parm     ,
 *         .    
 *      mular_descriptor cluster.
 *         
 *       babolo_parm    
 *  .
 *    .
 *          
 *    .     rawobj,   
 *  :
 *  - ,   popush_obj  mular_descriptor object   
 *    ;
 *  -  ,  ,  ,    
 *      popush_obj;
 *  -  popush_obj    (fdel .. frev)  mular_descriptor ask
 *        ;
 *  -      babolo_parm  rawobj 
 *    ask   mular_descriptor      
 *    fdel .. frev popush_obj,     ;
 *  -    ask,    .
 *    rawobj ,    popush_obj  
 *    .      
 *        rawobj.
 *  ask 0 .. 3    START, STOP, SIGNAL, SELF
 * ,  0  popush_obj.fdel .. popush_obj.frev 
 * , 1 .. 3  .
 *  START, STOP, SIGNAL ask  ,   
 *  popush_cnf.strt .. popush_cnf.sign,   SELF    
 *   .  ask  SELF ,   
 *   popush_cnf.self,   0 ().
 *  ask  SELF  ,  popushd     .
 */

extern u_int32_t popush_state;

popush_cnf *popush_syntan    __P((popush_main *main));
void        popush_execevent __P((popush_main *main));
int         popush_setevent  __P((popush_main *main));
int         popush_setkact   __P((popush_main *main, struct kevent *ev));
int         popush_diffmerge __P((popush_main *main, popush_cnf *cnf));
void        popush_setaction __P((popush_main *main, size_t a));
void        popush_setact    __P((popush_main *main, size_t a));
void        popush_cnfree    __P((popush_cnf *cfg));
